//
// LocoTool - A program to edit Locomotion object description .dat files
//
// Copyright (C) 2004 by Josef Drexler
//
// This file is distributed under the term of the GNU General Public Licence
// version 2 or higher.  Please read the file LICENCE for details.
//
//
// codec.c - converted between the RLE encoded .dat file and its decoded
//	     flat byte array; plus call the object class-specific handlers
//	     for encoding or decoding
//

#include <stdio.h>
#include <string.h>
#include <setjmp.h>
#include <unistd.h>
#include <sys/stat.h>

#include "common.h"
#include "locodef.h"
#include "dump.h"
#include "xml.h"
#include "locotool.h"

#ifdef __linux
#	define __declspec(a)
#	define mkdir(a) mkdir(a, 0777)
#endif
#ifdef __windows
#	define WIN32_LEAN_AND_MEAN
#	include <windows.h>
#endif

// Global variables

int embedpng = 0;	// can't be set atm, only 0 supported
int allinone = 0;	// can't be set atm, only 0 supported
int demofile = 0;	// if 1, generate checksum for the demo
int palettetype = 27;	// which colour palette to use for company colours
int quiet = 0;		// no output, no messages, fail silently
char *_filebase = NULL;	// filename with the extension removed
char *_pngbase = NULL;	// same, but as directory

#ifdef __windows
__declspec(dllexport) BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fwdReason, LPVOID lpvReserved)
{
/*
  FILE *log = fopen("locotool.dll.log", "at");
  fprintf(log, "locotool.dll DllMain(%p, %lx, %p) called\n", hinstDLL, fwdReason, lpvReserved);
  fclose(log);
*/
  return TRUE;
}
#endif

struct {
	char **s;
	int set;
} stroptions[opt_STRLAST] = {
	{ &_filebase },
	{ &_pngbase },
};

struct {
	int *v, set;
} intoptions[opt_INTLAST] = {
	{ &palettetype },
	{ &demofile },
	{ &quiet },
};

__declspec(dllexport) char *LocoTool_set_str_option(en_stroption opt, char *str)
{
  char *old;
  if (opt >= opt_STRLAST)
	die("String option %d out of range", opt);

  if (stroptions[opt].set == 2)
	die("Too late to set string option %d", opt);

  old = *stroptions[opt].s;
  *stroptions[opt].s = str;
  stroptions[opt].set = 1;

  return old;
}

__declspec(dllexport) char *LocoTool_get_str_option(en_stroption opt)
{
  if (opt >= opt_STRLAST)
	die("String option %d out of range", opt);

  return *stroptions[opt].s;
}

__declspec(dllexport) int LocoTool_set_int_option(en_stroption opt, int val)
{
  int old;

  if (opt >= opt_INTLAST)
	die("Integer option %d out of range", opt);

  old = *intoptions[opt].v;
  *intoptions[opt].v = val;
  intoptions[opt].set = 1;

  return old;
}

__declspec(dllexport) int LocoTool_get_int_option(en_stroption opt)
{
  if (opt >= opt_INTLAST)
	die("Integer option %d out of range", opt);

  return *intoptions[opt].v;
}

void clearoptions(void)
{
  if (stroptions[opt_filebase].set >= 2) {
	my_free(_filebase);
	stroptions[opt_filebase].set = 0;
  }
  if (stroptions[opt_pngbase].set >= 2) {
	my_free(_pngbase);
	stroptions[opt_pngbase].set = 0;
  }
}

char *makefilename(const char *suffix)
{
  char *name = my_alloc(strlen(*stroptions[opt_filebase].s)+strlen(suffix)+1, suffix);

  strcpy(name, *stroptions[opt_filebase].s);
  strcat(name, suffix);

  return name;
}

char *makepngname(int num)
{
  struct stat st;
  char *name;

  if (stroptions[opt_pngbase].set == 2) {
	if (stat(_pngbase, &st))
		if (mkdir(_pngbase))
			die("Cannot make directory %s: %s", _pngbase, strerror(errno));
#ifdef __linux
	  strcat(_pngbase, "/");
#else
	  strcat(_pngbase, "\\");
#endif

	stroptions[opt_pngbase].set = 3;
  }

  name = my_alloc(strlen(*stroptions[opt_pngbase].s)+16, "pngname");

  strcpy(name, *stroptions[opt_pngbase].s);
  sprintf(name+strlen(name), "%03d.png", num);

  return name;
}

// store the filename with the extension removed
void setfilebase(char *fname)
{
  if (!stroptions[opt_filebase].set) {
	_filebase = my_alloc(strlen(fname)+1, "filebase");

	strcpy(_filebase, fname);
	if (strrchr(_filebase, '.'))
		strrchr(_filebase, '.')[0] = 0;

	stroptions[opt_filebase].set = 2;
  }

  if (!stroptions[opt_pngbase].set) {
	_pngbase = my_alloc(strlen(_filebase)+2, "pngbase");

	strcpy(_pngbase, _filebase);

	stroptions[opt_pngbase].set = 2;
  }
}

// RLE decode the given data and return a malloc'ed decoded copy
// (we don't need to return the length because that's "obvious" from the
// contents of the data)
__declspec(dllexport) int LocoTool_rledecode(uint8_t *data, uint32_t len, uint8_t **outdata, uint32_t *outlen)
{
  uint8_t *chunk;
  uint32_t chunklen, chunkofs;

  chunklen = len*2;
  chunk = my_alloc(chunklen, "chunk");

  chunkofs = 0;
  while (len>0) {
	int8_t rle;
	int run;

	rle = *(int8_t*) data++; len--;
	run = abs(rle)+1;

	if (chunkofs+run > chunklen) {
		uint32_t oldlen = chunklen;
		while (chunkofs+run > chunklen)
			chunklen <<= 1;
		chunk = my_realloc(chunk, chunklen, oldlen, "chunk");
	}

	if (rle<0) {
		uint8_t byte = *data++; len--;

		memset(chunk+chunkofs, byte, run);
	} else {
		memcpy(chunk+chunkofs, data, run);
		len-=run;
		data+=run;
	}
	chunkofs += run;
  }

  *outdata = my_realloc(chunk, chunkofs, chunklen, "chunk");
  *outlen = chunkofs;

  return 1;
}

// RLE decode then string decompression
__declspec(dllexport) int LocoTool_decompress(uint8_t *data, uint32_t len, uint8_t **outdata, uint32_t *outlen)
{
  uint8_t *rle, *chunk;
  uint32_t rlelen, rleofs, chunklen, chunksize;

  if (!LocoTool_rledecode(data, len, &rle, &rlelen))
	return 0;

  chunksize = rlelen*2;
  chunk = my_alloc(chunksize, "chunk");

  chunklen = 0;
  rleofs = 0;
  while (rleofs < rlelen) {
	uint8_t code;
	uint32_t len, ofs;

	code = rle[rleofs++];

	if (chunklen+33 > chunksize) {
		uint32_t oldsize = chunksize;
		while (chunklen+33 > chunksize)
			chunksize <<= 1;
		chunk = my_realloc(chunk, chunksize, oldsize, "chunk");
	}

	if (code == 0xff) {
		chunk[chunklen++] = rle[rleofs++];
	} else {
		len = (code & 7)+1;
		ofs = 32 - (code >> 3);

		memcpy(chunk+chunklen, chunk+chunklen-ofs, len);
		chunklen+=len;
	}
  }

  my_free(rle);

  *outdata = my_realloc(chunk, chunklen, chunksize, "chunk");
  *outlen = chunklen;

  return 1;
}


// define a bit rotate (gcc optimizes this into an i386 ROL instruction)
#define _rotl32(a,n) (((a)<<(n))|((a)>>(32-n)))
#define _rotl8(a,n) (((a)<<(n))|((a)>>(8-n)))
#define _rotr32(a,n) (((a)>>(n))|((a)<<(32-n)))
#define _rotr8(a,n) (((a)>>(n))|((a)<<(8-n)))


// descramble using success bit rotates
__declspec(dllexport) int LocoTool_descramble(uint8_t *data, uint32_t len)
{
  uint8_t bits, d;
  uint32_t ofs;

  bits = 1;
  for (ofs=0; ofs<len; ofs++) {
	data[ofs] = _rotr8(data[ofs], bits);
	bits = (bits+2)&7;
  }

  return 1;
}

// RLE encode the given data, and return a malloc'ed encoded copy plus its
// length in the given parameters
__declspec(dllexport) int LocoTool_rleencode(uint8_t *data, uint32_t len, uint8_t **rleout, uint32_t *rlelen)
{
  uint8_t *out;
  uint32_t start, ofs, outlen, outofs;
  int run;
  uint8_t byte, rle;

  outlen = len*2;
  out = my_alloc(outlen, "out");

  ofs = 0;
  outofs = 0;
  while (len) {
	start = ofs;

	byte = data[ofs++];
	len--;

	run = 1;
	if (len && (byte == data[ofs])) {
		while (len && (byte == data[ofs]) && (run < 127)) {
			run++;
			ofs++;
			len--;
		}
		rle = -(run-1);
		out[outofs++] = rle;
		out[outofs++] = byte;
	} else {
		while (len && (byte != data[ofs]) && (run < 126)) {
			byte = data[ofs];
			run++;
			ofs++;
			len--;
		}
		if (len) {
			run--;
			ofs--;
			len++;
		}
		rle = run-1;
		out[outofs++] = rle;
		memcpy(out+outofs, data+start, run);
		outofs += run;
	}
  }

  *rlelen = outofs;
  *rleout = out;

  return 1;
}

// scramble using success bit rotates
__declspec(dllexport) int LocoTool_scramble(uint8_t *data, uint32_t len)
{
  uint8_t bits, d;
  uint32_t ofs;

  bits = 1;
  for (ofs=0; ofs<len; ofs++) {
	data[ofs] = _rotl8(data[ofs], bits);
	bits = (bits+2)&7;
  }

  return 1;
}

// calculate the checksum of the RLE decoded data
__declspec(dllexport) uint32_t LocoTool_checksum(uint8_t *data, uint32_t len)
{
  uint32_t i, c;

  c = 0xF369A75B;
  if (demofile) c = 0xE369A75B;

  // start with the class, skip the subclass and checksum (obviously)
  c = _rotl32(c ^ data[0], 11);
  for (i=4; i<12; i++)
	c = _rotl32(c ^ data[i], 11);

  // then iterate over all remaining bytes
  for (i=16; i<len; i++)
	c = _rotl32(c ^ data[i], 11);

  return c;
}

#ifdef DEBUGDMP
FILE *dmp;
extern int colhist[256];
#endif

// call the object class-specific dump handler after a sanity check
int dumpchunk(FILE *xml, uint8_t *data, uint32_t len, int class)
{
  uint32_t dumped;

#ifdef DEBUGDMP
  fwrite(data, 1, len, dmp);
  fflush(dmp);
#endif

  if ( (class >= NUMOBJCLASSES) || !objclasses[class].dump)
	die("Can't translate class %02X yet", class);

  dumped =  dumpvars(xml, data, objclasses[class].vars);
  dumped += objclasses[class].dump(xml, data, dumped);

  return dumped;
}

// Exported: decode the given file into an xml an a slew of png files
__declspec(dllexport) int LocoTool_decode(char *datname)
{
  int class;
  uint32_t subclass;
  uint8_t *chunk;
  uint32_t dumped;

  uint8_t *data, *ptr, *end;
  struct stat st;
  FILE *dat, *xml;
  char *xmlname;
  jmp_buf decode_err;

  // in case of an error, we return here with 0
  err_return = &decode_err;
  if (setjmp(*err_return)) {
	err_return = NULL;
	return 0;
  }

  setfilebase(datname);

  xmlname = makefilename(".xml");

  dat = fopen(datname, "rb");
  if (!dat) die("Can't read %s: %s", datname, strerror(errno));

  // get the file size
  if (fstat(fileno(dat), &st)) die("Can't stat %s: %s", datname, strerror(errno));

  // slurp the whole file
  data = my_alloc(st.st_size, "data");
  fread(data, 1, st.st_size, dat);
  fclose(dat);

  // sanity check that it's really a Locomotion .dat file
  if (data[3] != 0x11) die("File %s is not a .dat file", datname);
  end = data+st.st_size;

  class = data[0];
  subclass = getvalue(data+1, 3);

  if ( ((class & 0x7f) >= NUMOBJCLASSES) || !objclasses[class&0x7f].dump)
	die("Can't translate class %02X yet", class);

  if (data[16] > 3)
	die("Can't decode chunk compression %d", data[16]);

  xml = fopen(xmlname, "wt");
  if (!xml) die("Can't write %s: %s", xmlname, strerror(errno));
  my_free(xmlname);

  // dump the file contents as xml
  fputs("<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?>\n", xml);

  fprintf(xml, "<object class=\"0x%02X\" subclass=\"0x%06X\" name=\"",
	class, subclass);
  putxmlstr(xml, data+4, 8);
  fputs("\">", xml);

  ptr = data+16;

  class &= 0x7f;

  // For debugging the object data assembly functions
#ifdef DEBUGDMP
  {
    char *dmpname = makefilename(".dmp");
    dmp=fopen(dmpname,"wb"); if (!dmp) die("write %s: %s", dmpname, strerror(errno));
    my_free(dmpname);
    fwrite(data, 1, 16, dmp);
#endif

  while (ptr < end) {
	int compression;
	uint32_t len, chunklen;

	compression = *ptr; ptr++;
	len = *(uint32_t*) ptr; ptr+=4;

	fprintf(xml, "<chunk compression=\"%d\">\n", compression);
	switch(compression) {
		case 0:
			dumpchunk(xml, ptr, len, class);
			break;
		case 1:
			if (!LocoTool_rledecode(ptr, len, &chunk, &chunklen))
				die("Error while RLE decoding");
			dumpchunk(xml, chunk, chunklen, class);
			my_free(chunk);
			break;
		case 2:
			if (!LocoTool_decompress(ptr, len, &chunk, &chunklen))
				die("Error while RLE decompressing");
			dumpchunk(xml, chunk, chunklen, class);
			break;
		case 3:
			LocoTool_descramble(ptr, len);
			dumpchunk(xml, ptr, len, class);
			break;
		default:
			die("Can't decode chunk compression %d", compression);
	}

	fputs("</chunk>", xml);
	ptr += len;
  }
  fputs("</object>\n", xml);

#ifdef DEBUGDMP
    fclose(dmp);
    dmpname = makefilename(".col");
    dmp=fopen(dmpname,"wb"); if (!dmp) die("write %s: %s", dmpname, strerror(errno));
    my_free(dmpname);
    fwrite(colhist, 1, sizeof(colhist), dmp);
    fclose(dmp);
      }
#endif

  fclose(xml);

  my_free(data);

  clearoptions();

  // if an error happens later, we *don't* want to return to the above place!
  err_return = NULL;
  return 1;
}

// Exported: Encode the given xml file into a Locomotion .dat file
__declspec(dllexport) int LocoTool_encode(char *xmlname)
{
  int compression;
  uint32_t ofs, rlelen, chksum, encsize;
  struct stat st;

  FILE *xml, *dat;
  char *datname, *tmpname, *bakname;
  uint8_t *xmldata, *rle, *encdata;
  jmp_buf encode_err;

  // error handler, on error we return here with 0
  err_return = &encode_err;
  if (setjmp(*err_return)) {
	err_return = NULL;
	return 0;
  }

  setfilebase(xmlname);

  datname = makefilename(".dat");
  tmpname = makefilename(".tmp");
  bakname = makefilename(".bak");

  xml = fopen(xmlname, "rb");	// rt for real xml!!
  if (!xml) die("Can't read %s: %s", xmlname, strerror(errno));

  if (fstat(fileno(xml), &st)) die("Can't stat %s: %s", xmlname, strerror(errno));

  // slurp the whole xml file
  xmldata = my_alloc(st.st_size, "data");
  fread(xmldata, 1, st.st_size, xml);
  fclose(xml);

  // parse it
  encsize = parsexml(xmldata, st.st_size, &encdata, &compression);

  // write the file header after calculating the checksum
  dat = fopen(tmpname, "w+b");
  chksum = LocoTool_checksum(encdata, encsize);
  *(uint32_t*) (encdata+12) = chksum;
  fwrite(encdata, 1, 16, dat);

  // For debugging the object data assembly functions
#ifdef DEBUGDMP
  {
    FILE *f;
    char *dmx = makefilename(".dmx");
    f=fopen(dmx,"wb"); if (!f) die("write %s: %s", dmx, strerror(errno));
    fwrite(encdata, 1, encsize, f);
    fclose(f);
    my_free(dmx);
  }
#endif

	// cheat a bit: don't do string compression for now
  if (compression == 2)
	compression = 1;

  fwrite(&compression, 1, 1, dat);

  switch (compression) {
	case 0:
		// write the flat data structure
		rle = encdata+16;	// not really RLE encoded!
		rlelen = encsize-16;
		fwrite(&rlelen, 4, 1, dat);
		fwrite(rle, 1, rlelen, dat);
		break;
	case 1:
		// rle-encode the rest of the file
		if (!LocoTool_rleencode(encdata+16, encsize-16, &rle, &rlelen))
			die("Error while RLE encoding");
		fwrite(&rlelen, 4, 1, dat);
		fwrite(rle, 1, rlelen, dat);
		my_free(rle);
		break;
	case 3:
		LocoTool_scramble(encdata+16, encsize-16);
		rlelen = encsize-16;
		fwrite(&rlelen, 4, 1, dat);
		fwrite(encdata+16, 1, encsize-16, dat);
		break;
	default:
		die("Can't encode chunk compression %d", compression);
  }

  fclose(dat);

  // make a backup of the old .dat if not .bak exists yet
  if (!access(datname, 0)) {
	if (access(bakname, 0)) {
		rename(datname, bakname);
	} else {
		remove(datname);
	}
  }
  rename(tmpname, datname);

  my_free(encdata);
  my_free(xmldata);

  clearoptions();

  err_return = NULL;
  return 1;
}
