//
// LocoTool - A program to edit Locomotion object description .dat files
//
// Copyright (C) 2004 by Josef Drexler
//
// This file is distributed under the term of the GNU General Public Licence
// version 2 or higher.  Please read the file LICENCE for details.
//
//
// dump.c - dump object data structure entries into an xml file
//

#include <string.h>
#include <stdarg.h>

#include "common.h"
#include "dump.h"
#include "xml.h"
#include "png.h"
#include "locodef.h"

// up to 10 indentations with no silly loop to make tabs
char indent[11] = "\t\t\t\t\t\t\t\t\t\t";
#define INDENT (indent+10-indentation)

// read an unsigned value of the given size
uint32_t getuvalue(uint8_t *data, int size)
{
  uint32_t value = *(uint32_t*)(data);
  value &= 0xffffffff >> (32-8*size);

  return value;
}

// same but signed
int32_t getsvalue(uint8_t *data, int size)
{
  int32_t value = *(int32_t*)(data);
  value = (value << (32-8*size)) >> (32-8*size);

  return value;
}

// and automatic, unsigned for size>0 and signed for size<0
int32_t getvalue(uint8_t *data, int size)
{
  if (size > 0)
	return getuvalue(data, size);
  else
	return getsvalue(data, -size);
}


//
// ** dump handlers
//
// These dump various variables/structures/bit fields to the xml file
// They all return the number of bytes they ate
//

// write a bit field
uint32_t dumpflags(FILE *xml, uint8_t *data, const char **flags, int size, int indentation)
{
  uint32_t value, i, state;
  const char *name;
  char defname[32];

  if (size > 4) die("Can't dump flags with %d bytes", size);

  value = getvalue(data, size);

  for (i=0; i<size*8; i++) {
	if (*flags) {
		name = *flags;
		flags++;
	} else
		name = "";

	if (!strlen(name)) {
		sprintf(defname, "bit_%X", i);
		name = defname;
	}
	state = (value & (1 << i)) >> i;
	if ((name != defname) || state)
		fprintf(xml, "%s<bit name=\"%s\">%d</bit>\n", INDENT, name, state);
  }
  return size;
}

// dump a structure to the xml file, structure definition in the vars parameter
uint32_t dodump(FILE *xml, uint8_t *data, varinf *vars, int indentation)
{
  int i, j, size;
  uint32_t dumped, totaldumped, ofs;
  int32_t value;
  const char *name, *fname, *tag;
  char defname[32], arrname[32];

  totaldumped = 0;
  ofs = 0;

  for (i=0; vars[i].ofs != -1; i++) {
	fname = vars[i].name;
	if (!strlen(fname)) {
		sprintf(defname, "field_%X", vars[i].ofs);
		fname=defname;
	}
	if (ofs != vars[i].ofs) die("Structure is invalid at %X", ofs);
	for (j=0; j<vars[i].num; j++) {
		name = fname;
		if (vars[i].num > 1) {
			sprintf(arrname, "%s[%d]", fname, j);
			name = arrname;
		}

		size = vars[i].size;
		if (!size)
			die("Variable %s has no size", name);

		if (vars[i].flags) {
			fprintf(xml, "%s<bitmask name=\"%s\" size=\"%d\">\n", INDENT, name, size);
			dumped = dumpflags(xml, data+ofs, vars[i].flags, vars[i].size, indentation+1);
			fprintf(xml, "%s</bitmask>\n", INDENT);
		} else if (vars[i].structvars) {
			fprintf(xml, "%s<structure name=\"%s\" size=\"%d\">\n", INDENT, name, size);
			dumped = dodump(xml, data+ofs, vars[i].structvars,
				indentation+1);

			fprintf(xml, "%s</structure>\n", INDENT);
		} else {
			value = getvalue(data+ofs, size);
			size = abs(size);
			tag = "variable";
			if (!strlen(vars[i].name)) {
				if (value)
					tag = "unknown";
				else
					tag = NULL;	// don't print unset unknown variables
			}

			if (tag)
				fprintf(xml, "%s<%s name=\"%s\" size=\"%X\">%d</%s>\n", INDENT, tag, name, size, value, tag);
			dumped = size;
		}
		ofs += dumped;
		totaldumped += dumped;
	}
  }
  return totaldumped;
}

// this is the function actually called by the dump procs, all it does
// is dump the structure in the vars parameter with an initial indentation of 1
uint32_t dumpvars(FILE *xml, uint8_t *data, varinf *vars)
{
  return dodump(xml, data, vars, 1);
}

// dump the language/description info
// those are a bunch of zero-terminated strings preceeded by the language byte
uint32_t dumplang(FILE *xml, uint8_t *data, int num)
{
  int language;
  uint32_t ofs;

  ofs = 0;
  while ( (language = data[ofs++]) != 0xff ) {
	fprintf(xml, "\t<description num=\"%d\" language=\"%d\">", num, language);
	putxmlstr(xml, data+ofs, 0);
	fputs("</description>\n", xml);
	ofs += strlen(data+ofs) + 1;
  }

  return ofs;
}

// dump an object dependence (i.e. an outside object that this object depends on)
uint32_t dumpuseobj(FILE *xml, uint8_t *data, int num, int total, useobjid type, ...)
{
  const char *typename;
  char arrtype[32];
  int class, classok;
  va_list ap;

  typename = useobjnames[type];
  if (total > 1) {
	sprintf(arrtype, "%s[%d]", typename, num);
	typename = arrtype;
  }

  va_start(ap, type);

  classok = 0;
  while ( (class = va_arg(ap, int)) != -1)
	if ( (class == data[0]) || (class == 0x100) )
		classok = 1;
  va_end(ap);

  if (!classok)
	die("Not correct class for %s: %02X", typename, data[0]);

  fprintf(xml, "\t<useobject desc=\"%s\" class=\"%d\">%.8s</useobject>\n",
	typename, data[0], data+4);

  return 16;
}

// dump a cargo capacity entry with various cargo types
uint32_t dumpcap(FILE *xml, uint8_t *data, int num, int total)
{
  uint32_t dumped;
  int capacity, cargotype, refcap;

  if (total);

  capacity = data[0];

  dumped = 1;

  if (capacity) {
	fprintf(xml, "\t<cargo num=\"%d\" capacity=\"%d\">\n", num, capacity);
	while (getvalue(data+dumped, 2) != 0xffff) {
		cargotype = data[dumped++];

		refcap = getvalue(data+dumped, 2);
		dumped += 2;

		fprintf(xml, "\t\t<cargotype id=\"%d\">%d</cargotype>\n", cargotype, refcap);
	}
	dumped += 2;
	fputs("\t</cargo>\n", xml);
  }

  return dumped;
}

// dump the list of sprites plus their contents
uint32_t dumpsprites(FILE *xml, uint8_t *data)
{
  uint32_t i, j, num, size, totalsize, dumped;
  uint8_t *spritedata;
  FILE *png = NULL;

  num = getvalue(data, 4);
  size = getvalue(data+4, 4);
  dumped = 8;

  spritedata = data+8+16*num;
  totalsize = 0;

  for (i=0; i<num; i++) {
	uint32_t ofs,width,height,flags, spritesize;
	int32_t xofs,yofs;

	ofs=getvalue(data+dumped,4);  dumped+=4;
	width=getvalue(data+dumped,2);  dumped+=2;
	height=getvalue(data+dumped,2);  dumped+=2;
	xofs=getsvalue(data+dumped,2);  dumped+=2;
	yofs=getsvalue(data+dumped,2);  dumped+=2;
	flags=getvalue(data+dumped,4);  dumped+=4;

	fprintf(xml, "\t<sprite id=\"%d\" xofs=\"%d\" yofs=\"%d\">\n", i, xofs, yofs);
	dumpflags(xml, (uint8_t*) &flags, spriteflags, 4, 2);

	if ( (width==1) && (height==1) ) {
		uint8_t col;

		spritesize = putspriterow(spritedata+ofs, flags, 0, width, &col);

		fprintf(xml, "\t\t<stub>%d</stub>\n", col);
	} else if (flags & 0x40) {
		spritesize = 0;
	} else if (embedpng) {
		fputs("\t\t<pngdata>", xml);
		fputs("</pngdata>\n", xml);
                spritesize = 0;
	} else if (allinone) {
		// fprintf(xml, "\t\t<png x= y= width= height=>%s</png>\n", );
		spritesize = 0;
	} else {
		char *pngname = makepngname(i);
		png = fopen(pngname, "wb");
		if (!png) die("Can't write %s: %s", pngname, strerror(errno));

		fprintf(xml, "\t\t<pngfile>%s</pngfile>\n", pngname);

		my_free(pngname);

		spritesize = makepng(xml, png, spritedata + ofs, width, height, flags);
		fclose(png);

	}

	totalsize += spritesize;

	fputs("\t</sprite>\n", xml);
  }

  dumped += totalsize;
  if (totalsize != size) die("Sprite data different from size; wanted %d had %d", size, totalsize);

  return dumped;
}

varinf auxvars[] = {
	{ 0x00, 0, 0, "" },
	{ -1 },
};

uint32_t dumpaux(FILE *xml, uint8_t *data, int id, int size, int num)
{
  int i, type, siz;
  uint32_t dumped;

  type = 0;
  dumped = 0;
  siz = abs(size);
  if (num < 0) {
	type = -num;
	num = 0;
	for (i=0; getsvalue(data+i, type) != -1; i+=siz) num++;
	dumped+=type;
  }

  fprintf(xml, "\t<auxdata id=\"%d\" size=\"%d\" num=\"%d\" type=\"%d\">\n", id, siz, num, type);

  if (size < 0) {
	num *= -size;
	size = 1;
  }
  auxvars[0].size = size;
  auxvars[0].num = num;
  dumped += dodump(xml, data, auxvars, 2);

  fputs("\t</auxdata>\n", xml);

  return dumped;
}

uint32_t dumpstrtable(FILE *xml, uint8_t *data, int id, int num)
{
  int i;
  uint32_t dumped;

  dumped = num*2;
  fprintf(xml, "\t<stringtable id=\"%d\" num=\"%d\">\n", id, num);
  for (i=0; i<num; i++) {
	dumped = getvalue(data+i*2, 2);
	fprintf(xml, "\t\t<string id=\"%d\" type=\"%d\">", i, data[dumped+strlen(data+dumped)+1]);
	dumped += putxmlstr(xml, data+dumped, 0)+2;
	fputs("</string>\n", xml);
  }
  fputs("\t</stringtable>\n", xml);

  return dumped;
}
