//
// LocoTool - A program to edit Locomotion object description .dat files
//
// Copyright (C) 2004 by Josef Drexler
//
// This file is distributed under the term of the GNU General Public Licence
// version 2 or higher.  Please read the file LICENCE for details.
//
//
// objects.c - object class-specific handlers and data structure definitions
//

#include <stdlib.h>
#include <string.h>

#include "common.h"
#include "dump.h"
#include "locodef.h"
#include "xml.h"

//
// Two main contents of this file:
// - the object class-specific data structures
//   (in varinf arrays plsu char* arrays for the bit fields)
// - the object class-specific decode and encode handlers
//
// Each varinf entry is either a simple variable, a sub-structure, a
// bit field, or an array of those.  All must be in sequence without
// holes.
//
// The handlers really only decide on additional information to use, the
// main object data is handled through its varinf structure, but the order
// of external object definitions, descriptions and other items is also
// class specific and thus needs its own handler.
//
//
// All object classes are then put into the objclasses[] array at the bottom
// of this file.
//
// Since everything is about the same for all classes, only the vehicle
// class has useful comments, so look there for guidance.
//

// ***********************
//      General items
// ***********************

	// the known bits in the flags of each sprite
const char *spriteflags[] = {
	"hasdata", "", "chunked", "",
	"", "", "copy", "",
	NULL
};

// ***********************
//  SIMPLE CLASS HANDLER
// ***********************

varinf simplevars[] = {
	{ 0x00, 1, 6, "" },
	{ -1 },
};

uint32_t dumpsimple(FILE *xml, uint8_t *data, uint32_t dumped)
{
  dumped += dumplang(xml, data+dumped, 0);
  dumped += dumpsprites(xml, data+dumped);

  return dumped;
}

void parsesimple(xmlinf *inf)
{
  adddesc(inf, 0);
  addsprites(inf);
}

uint32_t dumpnosprites(FILE *xml, uint8_t *data, uint32_t dumped)
{
  dumped += dumplang(xml, data+dumped, 0);

  return dumped;
}

void parsenosprites(xmlinf *inf)
{
  adddesc(inf, 0);
}


// ***********************
// Class 00:  INTERFACES
// ***********************

varinf interfacevars[] = {
	{ 0x00, 1, 24, "" },
	{ -1 },
};


// ***********************
// Class 01:  SOUNDS
// ***********************

varinf sfxvars[] = {
	{ 0x00, 1, 12, "" },
	{ -1 },
};


uint32_t dumpsfx(FILE *xml, uint8_t *data, uint32_t dumped)
{
  dumped += dumplang(xml, data+dumped, 0);
/*
  dumped += dumpsounds(xml, data+dumped);
*/
  return dumped;
}

void parsesfx(xmlinf *inf)
{
/*
  adddesc(inf, 0);
  addsounds(inf);
  */
}


// ***********************
// Class 02:  CURRENCIES
// ***********************

varinf currvars[] = {
	{ 0x00, 1, 10, "" },
	{ 0x0A, 1, 1, "zeroes" },
	{ 0x0B, 1, 1, "shiftnum" },
	{ -1 },
};

uint32_t dumpcurr(FILE *xml, uint8_t *data, uint32_t dumped)
{
  dumped += dumplang(xml, data+dumped, 0);
  dumped += dumplang(xml, data+dumped, 1);
  dumped += dumplang(xml, data+dumped, 2);
  dumped += dumpsprites(xml, data+dumped);

  return dumped;
}

void parsecurr(xmlinf *inf)
{
  adddesc(inf, 0);
  adddesc(inf, 1);
  adddesc(inf, 2);
  addsprites(inf);
}



// ***********************
// Class 03:  EXHAUST FX
// ***********************

varinf exhfxvars[] = {
	{ 0x00, 1, 30, "" },
	{ 0x1E, 1, 1, "numsnd" },
	{ 0x1F, 1, 9, "" },
	{ -1 },
};

uint32_t dumpexhfx(FILE *xml, uint8_t *data, uint32_t dumped)
{
  int i;

  dumped += dumplang(xml, data+dumped, 0);
  dumped += dumpaux(xml, data+dumped, 0, 2, -1);
  dumped += dumpaux(xml, data+dumped, 1, 2, -1);
  for (i=0; i<data[0x1e]; i++)
	dumped += dumpuseobj(xml, data+dumped, i, data[0x1e], ob_soundeffect, 0x01, -1);
  dumped += dumpsprites(xml, data+dumped);

  return dumped;
}

void parseexhfx(xmlinf *inf)
{
  int i;

  adddesc(inf, 0);
  addaux(inf, 0);
  addaux(inf, 1);
  for (i=0; i<inf->objdata[0x1e]; i++)
	addobj(inf, i, ob_soundeffect, 0x1, -1);
  addsprites(inf);
}


// ***********************
// Class 04:  CLIFF FACES
// ***********************

// (see simple vars above)


// ***********************
// Class 05:   WATER
// ***********************

varinf watervars[] = {
	{ 0x00, 1, 14, "" },
	{ -1 },
};


// ***********************
// Class 06:  GROUND
// ***********************

varinf groundvars[] = {
	{ 0x00, 1, 30, "" },
	{ -1 },
};

uint32_t dumpground(FILE *xml, uint8_t *data, uint32_t dumped)
{
  dumped += dumplang(xml, data+dumped, 0);
  dumped += dumpuseobj(xml, data+dumped, 0, 0, ob_cliff, 0x04, -1);
  dumped += dumpsprites(xml, data+dumped);

  return dumped;
}

void parseground(xmlinf *inf)
{
  adddesc(inf, 0);
  addobj(inf, 0, ob_cliff, 0x04, -1);
  addsprites(inf);
}


// ***********************
// Class 07:  TOWN NAMES
// ***********************

varinf townpart[] = {
	{ 0x00, 1, 1, "num" },
	{ 0x01, 1, 1, "numempty" },
	{ 0x02, 2, 1, "indexofs" },
	{ -1 },
};

varinf townvars[] = {
	{ 0x00, 1, 2, "" },
	{ 0x02, 4, 6, "part", townpart },
	{ -1 },
};

uint32_t dumptownnames(FILE *xml, uint8_t *data, uint32_t dumped)
{
  int i;

  dumped += dumplang(xml, data+dumped, 0);
  for (i=0; i<6; i++)
	dumped += dumpstrtable(xml, data+dumped, i, data[2+4*i]);
  return dumped;
}

void parsetownnames(xmlinf *inf)
{
  int i;
  uint32_t ofs;

  adddesc(inf, 0);
  for (i=0; i<6; i++) {
	// set offset
	if (inf->objdata[2+i*4])
		*(uint16_t*) (inf->objdata+4+i*4) = inf->objdata_i.ent;
	addstrtable(inf, i);
  }
}


// ***********************
// Class 08:  CARGO TYPES
// ***********************

const char *cargoflags[] = {
	"", "refitoption",
	NULL,
};

varinf cargovars[] = {
	{ 0x00, 1, 2, "" },
	{ 0x02, 2, 1, "unitweight" },
	{ 0x04, 1, 12, "" },
	{ 0x10, 1, 1, "id" },
	{ 0x11, 1, 1, "" },
	{ 0x12, 1, 1, "flags", NULL, cargoflags },
	{ 0x13, 1, 2, "" },
	{ 0x15, 1, 1, "peakdays" },
	{ 0x16, 1, 1, "decay1days" },
	{ 0x17, 2, 1, "decay1rate" },
	{ 0x19, 2, 1, "decay2rate" },
	{ 0x1B, 2, 1, "paymentfactor" },
	{ 0x1D, 1, 1, "paymentind" },
	{ 0x1E, 1, 1, "" },
	{ -1 },
};

uint32_t dumpcargo(FILE *xml, uint8_t *data, uint32_t dumped)
{
  dumped += dumplang(xml, data+dumped, 0);
  dumped += dumplang(xml, data+dumped, 1);
  dumped += dumplang(xml, data+dumped, 2);
  dumped += dumplang(xml, data+dumped, 3);
  dumped += dumpsprites(xml, data+dumped);

  return dumped;
}

void parsecargo(xmlinf *inf)
{
  adddesc(inf, 0);
  adddesc(inf, 1);
  adddesc(inf, 2);
  adddesc(inf, 3);
  addsprites(inf);
}


// ***********************
// Class 09:    FENCES
// ***********************

const char *fenceflags[] = {
	NULL
};

varinf fencevars[] = {
	{ 0x00, 1, 7, "" },
	{ 0x07, 1, 1, "flags", NULL, fenceflags },
	{ 0x08, 1, 1, "" },
	{ 0x09, 1, 1, "" },
	{ -1 },
};


// ***********************
// Class 0A:    SIGNALS
// ***********************

varinf signalvars[] = {
	{ 0x00, 1, 30, "" },
	{ -1 },
};

uint32_t dumpsignals(FILE *xml, uint8_t *data, uint32_t dumped)
{
  dumped += dumplang(xml, data+dumped, 0);
  dumped += dumplang(xml, data+dumped, 1);
  dumped += dumpsprites(xml, data+dumped);

  return dumped;
}

void parsesignals(xmlinf *inf)
{
  adddesc(inf, 0);
  adddesc(inf, 1);
  addsprites(inf);
}


// ***********************
// Class 0B:  CROSSINGS
// ***********************

varinf crossingvars[] = {
	{ 0x00, 1, 18, "" },
	{ -1 },
};


// ***********************
// Class 0C: STREET LIGHTS
// ***********************

varinf lightvars[] = {
	{ 0x00, 1, 12, "" },
	{ -1 },
};


// ***********************
// Class 0D:    TUNNELS
// ***********************

// (see simplevars above)


// ***********************
// Class 0E:   BRIDGES
// ***********************

varinf bridgevars[] = {
	{ 0x00, 1, 2, "" },
	{ 0x02, 1, 1, "noroof" },
	{ 0x03, 1, 1, "" },
	{ 0x04, 1, 1, "" },
	{ 0x05, 1, 1, "" },
	{ 0x06, 1, 1, "" },
	{ 0x07, 1, 1, "" },
	{ 0x08, 1, 1, "spanlength" },
	{ 0x09, 1, 1, "pillarspacing" },
	{ 0x0A, 2, 1, "maxspeed" },
	{ 0x0C, 1, 1, "maxheight" },
	{ 0x0D, 1, 1, "costind" },
	{ 0x0E,-2, 1, "basecostfact" },
	{ 0x10,-2, 1, "heightcostfact" },
	{ 0x12,-2, 1, "salecostfact" },
	{ 0x14, 2, 1, "disabledtrackcfg" },
	{ 0x16, 1,22, "" },
	{ -1 },
};


// ***********************
// Class 0F:  TRAIN STATIONS
// ***********************

varinf trnstatvars[] = {
	{ 0x00, 1, 34, "" },
	{ 0x22, 1, 1, "numcompat" },
	{ 0x23, 1, 7, "" },
	{ 0x2a, 1, 2, "designed" },
	{ 0x2c, 1, 2, "obsolete" },
	{ 0x2e, 1, 128, "" },
	{ -1 },
};

uint32_t dumptrnstat(FILE *xml, uint8_t *data, uint32_t dumped)
{
  int i;

  dumped += dumplang(xml, data+dumped, 0);
  for (i=0; i<32; i++)
	dumped += dumpaux(xml, data+dumped, i, 1, -4);
  dumped += dumpsprites(xml, data+dumped);

  return dumped;
}

void parsetrnstat(xmlinf *inf)
{
  int i;

  adddesc(inf, 0);
  for (i=0; i<32; i++)
	addaux(inf, i);
  addsprites(inf);
}


// ***********************
// Class 10:  TRACK MODS
// ***********************

varinf trkmodvars[] = {
	{ 0x00, 1, 18, "" },
	{ -1 },
};


// ***********************
// Class 11:   TRACKS
// ***********************

const char *track_curvetypes[] = {
	"", "widecurve", "mediumcurve", "smallcurve",
	"tightcurve", "normalslope", "wideslope",
	NULL,
};

varinf trackvars[] = {
	{ 0x00, 1, 2, "" },
	{ 0x02, 1, 1, "curvetypes", NULL, track_curvetypes },
	{ 0x03, 1, 4, "" },
	{ 0x07, 1, 1, "numcompat" },
	{ 0x08, 1, 1, "nummods" },
	{ 0x09, 1, 1, "numsignals" },
	{ 0x0A, 1, 10, "" },
	{ 0x14,-2, 1, "buildcostfact" },
	{ 0x16,-2, 1, "sellcostfact" },
	{ 0x18,-2, 1, "tunnelcostfact" },
	{ 0x1A, 1, 1, "costind" },
	{ 0x1B, 1, 1, "" },
	{ 0x1C, 2, 1, "curvespeed" },
	{ 0x1E, 1, 6, "" },
	{ 0x24, 1, 1, "numbridges" },
	{ 0x25, 1, 7, "" },
	{ 0x2C, 1, 1, "numstations" },
	{ 0x2D, 1, 7, "" },
	{ 0x34, 1, 1, "displayoffset" },
	{ 0x35, 1, 1, "" },
	{ -1 },
};

uint32_t dumptrack(FILE *xml, uint8_t *data, uint32_t dumped)
{
  int i;

  dumped += dumplang(xml, data+dumped, 0);
  for (i=0; i<data[7]; i++)
	dumped += dumpuseobj(xml, data+dumped, i, data[7], ob_compatible, 0x11, 0x14, -1);
  for (i=0; i<data[8]; i++)
	dumped += dumpuseobj(xml, data+dumped, i, data[8], ob_trackmod, 0x10, -1);
  for (i=0; i<data[9]; i++)
	dumped += dumpuseobj(xml, data+dumped, i, data[9], ob_signal, 0x0A, -1);
  dumped += dumpuseobj(xml, data+dumped, i, 0, ob_tunnel, 0x0D, -1);
  for (i=0; i<data[0x24]; i++)
	dumped += dumpuseobj(xml, data+dumped, i, data[0x24], ob_bridge, 0x0E, -1);
  for (i=0; i<data[0x2C]; i++)
	dumped += dumpuseobj(xml, data+dumped, i, data[0x2C], ob_station, 0x0F, -1);
  dumped += dumpsprites(xml, data+dumped);

  return dumped;
}

void parsetrack(xmlinf *inf)
{
  int i;

  adddesc(inf, 0);
  for (i=0; i<inf->objdata[7]; i++)
	addobj(inf, i, ob_compatible, 0x11, 0x14, -1);
  for (i=0; i<inf->objdata[8]; i++)
	addobj(inf, i, ob_trackmod, 0x10, -1);
  for (i=0; i<inf->objdata[9]; i++)
	addobj(inf, i, ob_signal, 0x0A, -1);
  addobj(inf, 0, ob_tunnel, 0x0D, -1);
  for (i=0; i<inf->objdata[0x24]; i++)
	addobj(inf, i, ob_bridge, 0x0E, -1);
  for (i=0; i<inf->objdata[0x2C]; i++)
	addobj(inf, i, ob_station, 0x0F, -1);
  addsprites(inf);
}


// ***********************
// Class 12:  ROAD STATIONS
// ***********************

varinf roadstvars[] = {
	{ 0x00, 1, 33, "" },
	{ 0x21, 1, 1, "numcompat" },
	{ 0x22, 1, 6, "" },
	{ 0x28, 2, 1, "designed" },
	{ 0x2a, 2, 1, "obsolete" },
	{ 0x2c, 1, 66, "" },
	{ -1 },
};

uint32_t dumproadst(FILE *xml, uint8_t *data, uint32_t dumped)
{
  int i;

  dumped += dumplang(xml, data+dumped, 0);
  dumped += dumpuseobj(xml, data+dumped, 0, 0, ob_cargo, 0x08, -1);
  for (i=0; i<16; i++)
	dumped += dumpaux(xml, data+dumped, i, 1, -4);
  dumped += dumpsprites(xml, data+dumped);

  return dumped;
}

void parseroadst(xmlinf *inf)
{
  int i;

  adddesc(inf, 0);
  addobj(inf, 0, ob_cargo, 0x08, -1);
  for (i=0; i<16; i++)
	addaux(inf, i);
  addsprites(inf);
}


// ***********************
// Class 13:  ROAD MODS
// ***********************

// (same as track mods)


// ***********************
// Class 14:    ROADS
// ***********************

varinf roadvars[] = {
	{ 0x00, 1, 20, "" },
	{ 0x14, 1, 1, "numbridges" },
	{ 0x15, 1, 8, "" },
	{ 0x1D, 1, 1, "numstations" },
	{ 0x1E, 1, 10, "" },
	{ 0x28, 1, 1, "numcompat" },
	{ 0x29, 1, 7, "" },
	{ -1 },
};

uint32_t dumproad(FILE *xml, uint8_t *data, uint32_t dumped)
{
  int i;

  dumped += dumplang(xml, data+dumped, 0);
  for (i=0; i<data[40]; i++)
	dumped += dumpuseobj(xml, data+dumped, i, data[40], ob_compatible, 0x11, 0x14, -1);
  for (i=0; i<data[37]; i++)
	dumped += dumpuseobj(xml, data+dumped, i, data[37], ob_roadmod, 0x13, -1);
  dumped += dumpuseobj(xml, data+dumped, i, 0, ob_tunnel, 0x0D, -1);
  for (i=0; i<data[20]; i++)
	dumped += dumpuseobj(xml, data+dumped, i, data[20], ob_bridge, 0x0E, -1);
  for (i=0; i<data[28]; i++)
	dumped += dumpuseobj(xml, data+dumped, i, data[28], ob_station, 0x12, -1);
  dumped += dumpsprites(xml, data+dumped);

  return dumped;
}

void parseroad(xmlinf *inf)
{
  int i;

  adddesc(inf, 0);
  for (i=0; i<inf->objdata[40]; i++)
	addobj(inf, i, ob_compatible, 0x11, 0x14);
  for (i=0; i<inf->objdata[37]; i++)
	addobj(inf, i, ob_roadmod, 0x13, -1);
  addobj(inf, 0, ob_tunnel, 0x0D, -1);
  for (i=0; i<inf->objdata[20]; i++)
	addobj(inf, i, ob_bridge, 0x0E, -1);
  for (i=0; i<inf->objdata[28]; i++)
	addobj(inf, i, ob_station, 0x12, -1);
  addsprites(inf);
}


// ***********************
// Class 15:   AIRPORTS
// ***********************

varinf airportvars[] = {
	{ 0x00, 1, 16, "" },
	{ 0x10, 2, 1, "allowedplanetypes" },
	{ 0x12, 1, 1, "numdata" },
	{ 0x13, 1, 149, "" },
	{ 0xA8, 2, 1, "designed" },
	{ 0xAA, 2, 1, "obsolete" },
	{ 0xAC, 1, 14, "" },
	{ -1 },
};

uint32_t dumpairport(FILE *xml, uint8_t *data, uint32_t dumped)
{
  int i, auxbase;

  dumped += dumplang(xml, data+dumped, 0);

  auxbase = 0;
  dumped += dumpaux(xml, data+dumped, auxbase++, 1, data[0x12]);
  dumped += dumpaux(xml, data+dumped, auxbase++, -2, data[0x12]);
  for (i=0; i<data[0x13]; i++)
	dumped += dumpaux(xml, data+dumped, auxbase++, 1, -1);
  dumped += dumpaux(xml, data+dumped, auxbase++, -4, -1);
  dumped += dumpaux(xml, data+dumped, auxbase++, 1, data[0xac]*8);
  dumped += dumpaux(xml, data+dumped, auxbase++, 1, data[0xad]*12);
  dumped += dumpsprites(xml, data+dumped);
  return dumped;
}

void parseairport(xmlinf *inf)
{
  int i, auxnum;

  adddesc(inf, 0);

  auxnum = inf->objdata[0x13]+5;
  for (i=0; i<auxnum; i++)
	addaux(inf, i);
  addsprites(inf);
}


// ***********************
// Class 16:     DOCKS
// ***********************

varinf dockvars[] = {
	{ 0x00, 1, 18, "" },
	{ 0x12, 1, 1, "numdata" },
	{ 0x13, 1, 13, "" },
	{ 0x20, 2, 1, "designed" },
	{ 0x22, 2, 1, "obsolete" },
	{ 0x24, 1, 4, "" },
	{ -1 },
};

uint32_t dumpdocks(FILE *xml, uint8_t *data, uint32_t dumped)
{
  int i, auxbase;

  dumped += dumplang(xml, data+dumped, 0);

  auxbase = 0;
  dumped += dumpaux(xml, data+dumped, auxbase++, 1, data[0x12]);
  dumped += dumpaux(xml, data+dumped, auxbase++, -2, data[0x12]);
  for (i=0; i<data[0x13]; i++)
	dumped += dumpaux(xml, data+dumped, auxbase++, 1, -1);
  dumped += dumpsprites(xml, data+dumped);
  return dumped;
}

void parsedocks(xmlinf *inf)
{
  int i, auxnum;

  adddesc(inf, 0);

  auxnum = inf->objdata[0x13]+2;
  for (i=0; i<auxnum; i++)
	addaux(inf, i);
  addsprites(inf);
}


// ***********************
// Class 17:   VEHICLES
// ***********************


	// the known bits in the vehicle sprite structure flags
const char *vehspriteflags[] = {
	"hassprites", "", "", "", "",
	"reversed",
	NULL
};

	// the vehicle sprite definition structure
varinf vehsprites[] = {
	{ 0x00, 1, 1, "numdir" },
	{ 0x01, 1, 1, "" },
	{ 0x02, 1, 1, "" },
	{ 0x03, 1, 1, "vehtype" },
	{ 0x04, 1, 1, "numunits" },
	{ 0x05, 1, 1, "" },
	{ 0x06, 1, 1, "bogeypos" },
	{ 0x07, 1, 1, "flags", NULL, vehspriteflags },
	{ 0x08, 1, 1, "" },
	{ 0x09, 1, 1, "" },
	{ 0x0A, 1, 1, "" },
	{ 0x0B, 1, 1, "" },
	{ 0x0C, 1, 1, "" },
	{ 0x0D, 1, 1, "" },
	{ 0x0E, 1, 1, "spritenum" },
	{ 0x0F, 1, 3, "" },
	{ 0x12, 4, 1, "" },
	{ 0x16, 4, 1, "" },
	{ 0x1A, 4, 1, "" },
	{ -1 },
};

	// a vehicle structure whose purpose is mostly unknown
varinf vehunk[] = {
	{ 0x00, 1, 1, "length" },
	{ 0x01, 1, 1, "" },
	{ 0x02, 1, 1, "" },
	{ 0x03, 1, 1, "" },
	{ 0x04, 1, 1, "spriteind" },
	{ 0x05, 1, 1, "" },
	{ -1 },
};

	// the vehicle flags
const char *vehflags[] = {
	"", "", "", "",
	"", "", "rackrail", "",
	"", "anytrack", "", "cancouple",
	"dualhead", "", "refittable", "noannounce",
	NULL
// note :: bit_B in the flags = disable coupling
};

	// the main vehicle data structure
varinf vehvars[] = {
	{ 0x000, 1, 2, "" },
	{ 0x002, 1, 1, "class" },
	{ 0x003, 1, 1, "type" },
	{ 0x004, 1, 1, "" },
	{ 0x005, 1, 1, "" },
	{ 0x006, 1, 1, "nummods" },
	{ 0x007, 1, 1, "costind" },
	{ 0x008,-2, 1, "costfact" },	// size -2 means signed word
	{ 0x00A, 1, 1, "reliability" },
	{ 0x00B, 1, 1, "runcostind" },
	{ 0x00C,-2, 1, "runcostfact" },
	{ 0x00E, 1, 1, "" },
	{ 0x00F, 1, 1, "numcompat" },
	{ 0x010, 1,20, "" },
	{ 0x024, 6, 4, "", vehunk },
	{ 0x03C,30, 4, "sprites", vehsprites },
	{ 0x0B4, 1,36, "" },
	{ 0x0D8, 2, 1, "power" },
	{ 0x0DA, 2, 1, "speed" },
	{ 0x0DC, 2, 1, "rackspeed" },
	{ 0x0DE, 2, 1, "weight" },
	{ 0x0E0, 2, 1, "flags", NULL, vehflags },
	{ 0x0E2, 1,44, "" },
	{ 0x10E, 1, 1, "visfxheight" },
	{ 0x10F, 1, 1, "visfxtype" },
	{ 0x110, 1, 1, "" },
	{ 0x111, 1, 1, "" },
	{ 0x112, 1, 1, "wakefxtype" },
	{ 0x113, 1, 1, "" },
	{ 0x114, 2, 1, "designed" },
	{ 0x116, 2, 1, "obsolete" },
	{ 0x118, 1, 1, "" },
	{ 0x119, 1, 1, "startsndtype" },
	{ 0x11A, 1,64, "" },
	{ 0x15A, 1, 1, "numsnd" },
	{ 0x15B, 1, 3, "" },
	{ -1 },
};

	// dump the vehicle data to the xml file
uint32_t dumpveh(FILE *xml, uint8_t *data, uint32_t dumped)
{
  int i;

	// the vehicle data structure has already been dumped,
	// so continue with the descriptions in various languages
  dumped += dumplang(xml, data+dumped, 0);

	// rail and road vehicles may be bound to specific track types
  if ((data[2] < 2) && !(getvalue(data+0xE0, 2) & 0x200))
	dumped += dumpuseobj(xml, data+dumped, 0, 0, ob_tracktype, 0x11 + data[2]*3, -1);

	// and they may require certain track modifications
  for (i=0; i<data[6]; i++)
	dumped += dumpuseobj(xml, data+dumped, i, data[6], ob_trackmod, 0x10 + data[2]*3, -1);

	// plus they have up to two cargo bays
  for (i=0; i<2; i++)
	dumped += dumpcap(xml, data+dumped, i, 2);

	// as well as a visual effect (e.g. steam/smoke/sparks)
  if (data[0x10F])
	dumped += dumpuseobj(xml, data+dumped, 0, 0, ob_visualeffect, 0x03, -1);

	// and for boats, a wake
  if (data[0x112])
	dumped += dumpuseobj(xml, data+dumped, 0, 0, ob_wakeeffect, 0x03, -1);

	// trains may use a rack rail modification
  if ((data[2] < 2) && (data[0xE0]&0x40))
	dumped += dumpuseobj(xml, data+dumped, 0, 0, ob_rackrail, 0x10, -1);

	// vehicles with which this one is compatible
	// (i.e. they can be in the same train)
  for (i=0; i<data[15]; i++)
	dumped += dumpuseobj(xml, data+dumped, i, data[15], ob_compatible, 0x17, -1);

	// start sound effect
  if (data[0x119])
	dumped += dumpuseobj(xml, data+dumped, 0, 0, ob_startsnd, 0x01, -1);

	// other sound effects
  for (i=0; i<(data[0x15A] & 0x7f); i++)
	dumped += dumpuseobj(xml, data+dumped, i, data[0x15A] & 0x7f, ob_soundeffect, 0x01, -1);

	// and finally the sprites
  dumped += dumpsprites(xml, data+dumped);

  return dumped;
}

	// assemble the vehicle data back out of the parsed xml info
void parseveh(xmlinf *inf)
{
  int i;
  uint8_t zero;

  zero = 0;

  // the vehicle structure is already assembled in the inf->objdata array
  // we only need to assemble the rest of the items

  adddesc(inf, 0);
  if ((inf->objdata[2] < 2) && !(getvalue(inf->objdata+0xE0, 2) & 0x200))
	addobj(inf, 0, ob_tracktype, 0x11 + inf->objdata[2]*3, -1);
  for (i=0; i<inf->objdata[6]; i++)
	addobj(inf, i, ob_trackmod, 0x10 + inf->objdata[2]*3, -1);
  for (i=0; i<2; i++)
	if (inf->capsize[i])
		addval(inf, inf->caps[i], inf->caplen[i]);
	else
		addval(inf, &zero, 1);
  if (inf->objdata[0x10F])
	addobj(inf, 0, ob_visualeffect, 0x03, -1);
  if (inf->objdata[0x112])
	addobj(inf, 0, ob_wakeeffect, 0x03, -1);
  if ((inf->objdata[2] < 2) && (inf->objdata[0xE0]&0x40))
	addobj(inf, 0, ob_rackrail, 0x10, -1);
  for (i=0; i<inf->objdata[15]; i++)
	addobj(inf, i, ob_compatible, 0x17, -1);
  if (inf->objdata[0x119])
	addobj(inf, 0, ob_startsnd, 0x01, -1);
  for (i=0; i<(inf->objdata[0x15A] & 0x7f); i++)
	addobj(inf, i, ob_soundeffect, 0x01, -1);
  addsprites(inf);
}

// ***********************
// Class 18:    TREES
// ***********************

varinf treevars[] = {
	{ 0x00, 1, 3, "" },
	{ 0x03, 1, 1, "height" },
	{ 0x04, 1, 72, "" },
	{ -1 },
};


// ***********************
// Class 19:    SNOW
// ***********************

// (see simple vars above)


// ***********************
// Class 1A:   CLIMATES
// ***********************

varinf climvars[] = {
	{ 0x00, 1, 10, "" },
	{ -1 },
};


// ***********************
// Class 1B:   SHAPES
// ***********************

varinf shapevars[] = {
	{ 0x00, 1, 14, "" },
	{ -1 },
};


// ***********************
// Class 1C:   BUILDINGS
// ***********************

varinf buildingvars[] = {
	{ 0x00, 1, 148, "" },
	{ 0x94, 2, 1, "earliestyr" },
	{ 0x96, 2, 1, "latestyr" },
	{ 0x98, 1, 38, "" },
	{ -1 },
};

uint32_t dumpbuildng(FILE *xml, uint8_t *data, uint32_t dumped)
{
  int i, num;

  dumped += dumplang(xml, data+dumped, 0);
  for (i=0; i<data[7]; i++)
	dumped += dumpaux(xml, data+dumped, i, 1, -1);
  for (i=0; i<4; i++)
	dumped += dumpuseobj(xml, data+dumped, i, 4, ob_cargo, 0x08, 0xFF, -1);
  for (i=0; i<data[0xad]; i++) {
	num = getvalue(data+dumped, 2);
	dumped += 2;
	dumped += dumpaux(xml, data+dumped, i+data[7], 1, num);
  }
  dumped += dumpsprites(xml, data+dumped);

  return dumped;
}

void parsebuildng(xmlinf *inf)
{
  int i, j, num, aux1;

  adddesc(inf, 0);
  aux1=inf->objdata[7];
  for (i=0; i<aux1; i++)
	addaux(inf, i);
  for (i=0; i<4; i++)
	addobj(inf, i, ob_cargo, 0x08, 0xFF, -1);
  for (i=0; i<inf->objdata[0xad]; i++) {
	num = -1;
	for (j=0; j<inf->auxdata_i.ent; j++)
		if (inf->auxdata[j]->id == i+aux1) {
			num = inf->auxdata[j]->num;
			break;
		}

	add16(inf, num);
	addaux(inf, i+aux1);
  }
  addsprites(inf);
}


// ***********************
// Class 1D:  SCAFFOLDING
// ***********************

varinf scaffvars[] = {
	{ 0x00, 1, 18, "" },
	{ -1 },
};


// ***********************
// Class 1E:  INDUSTRIES
// ***********************

const char *industryflags[] = {
	"", "", "", "",		// 1, 2, 4, 8
	"", "", "", "",		// 10, 20, 40, 80
	"", "", "", "",		// 100, 200, 400, 800
	"", "", "", "",		// 1000, 2000, 4000, 8000
	"", "needall", "canincreaseproduction", "candecreaseproduction",
				// 100000, 200000, 400000, 800000
	NULL,
};

varinf indvars[] = {
	{ 0x00, 1, 202, "" },
	{ 0xCA, 2, 1, "firstyear" },
	{ 0xCC, 2, 1, "lastyear" },
	{ 0xCE, 1, 1, "" },
	{ 0xCF, 1, 1, "costind" },
	{ 0xD0, 2, 1, "costfactor1" },
	{ 0xD2, 1, 18, "" },
	{ 0xE4, 4, 1, "flags", NULL, industryflags },
	{ 0xE8, 1, 12, "" },
	{ -1 },
};

uint32_t dumpind(FILE *xml, uint8_t *data, uint32_t dumped)
{
  int i;

  for (i=0; i<8; i++)
	dumped += dumplang(xml, data+dumped, i);
  dumped += dumpaux(xml, data+dumped, 0, 1, data[30]);
  dumped += dumpaux(xml, data+dumped, 1, 2, data[30]);
  dumped += dumpaux(xml, data+dumped, 2, 1, 4);
  dumped += dumpaux(xml, data+dumped, 3, 2, -1);
  for (i=0; i<data[31]; i++)
	dumped += dumpaux(xml, data+dumped, 4+i, 1, -1);
  dumped += dumpaux(xml, data+dumped, 4+data[31], 1, data[0xbd]);
  for (i=0; i<2; i++)
	dumped += dumpuseobj(xml, data+dumped, i, 2, ob_produces, 0x08, 0xff, -1);
  for (i=0; i<3; i++)
	dumped += dumpuseobj(xml, data+dumped, i, 3, ob_accepts, 0x08, 0xff, -1);
  for (i=0; i<6; i++)
	dumped += dumpuseobj(xml, data+dumped, i, 6, ob_fence, 0x09, 0xff, -1);
  dumped += dumpsprites(xml, data+dumped);

  return dumped;
}

void parseind(xmlinf *inf)
{
  int i;

  for (i=0; i<8; i++)
	adddesc(inf, i);
  addaux(inf, 0);
  addaux(inf, 1);
  addaux(inf, 2);
  addaux(inf, 3);
  for (i=0; i<inf->objdata[31]; i++)
	addaux(inf, 4+i);
  addaux(inf, 4+inf->objdata[31]);
  for (i=0; i<2; i++)
	addobj(inf, i, ob_produces, 0x08, 0xff, -1);
  for (i=0; i<3; i++)
	addobj(inf, i, ob_accepts, 0x08, 0xff, -1);
  for (i=0; i<6; i++)
	addobj(inf, i, ob_fence, 0x09, 0xff, -1);
  addsprites(inf);
}


// ***********************
// Class 1F:   REGIONS
// ***********************

varinf regionvars[] = {
	{ 0x00, 1, 18, "" },
	{ -1 },
};

uint32_t dumpregion(FILE *xml, uint8_t *data, uint32_t dumped)
{
  int i;

  dumped += dumplang(xml, data+dumped, 0);
  for (i=0; i<data[8]; i++)
	dumped += dumpuseobj(xml, data+dumped, i, data[8], ob_cargo, 0x08, -1);
  for (i=0; data[dumped] != 0xff; i++)
	dumped += dumpuseobj(xml, data+dumped, i, 0xffffff, ob_default, 0x100, -1);
  dumped++;
  dumped += dumpsprites(xml, data+dumped);

  return dumped;
}

void parseregion(xmlinf *inf)
{
  int i, numdef;

  adddesc(inf, 0);
  for (i=0; i<inf->objdata[8]; i++)
	addobj(inf, i, ob_cargo, 0x08, -1);
  numdef = inf->useobj_i.ent-inf->objdata[8];
  for (i=0; i<numdef; i++)
	addobj(inf, i, ob_default, 0x100, -1);
  add8(inf, 0xff);
  addsprites(inf);
}


// ***********************
// Class 20:  COMPANIES
// ***********************

varinf compvars[] = {
	{ 0x00, 1, 56, "" },
	{ -1 },
};

uint32_t dumpcomp(FILE *xml, uint8_t *data, uint32_t dumped)
{
  dumped += dumplang(xml, data+dumped, 0);
  dumped += dumplang(xml, data+dumped, 1);
  dumped += dumpsprites(xml, data+dumped);

  return dumped;
}

void parsecomp(xmlinf *inf)
{
  adddesc(inf, 0);
  adddesc(inf, 1);
  addsprites(inf);
}


// ***********************
// Class 21:   TEXTS
// ***********************

// (uses simple vars)

uint32_t dumptexts(FILE *xml, uint8_t *data, uint32_t dumped)
{
  dumped += dumplang(xml, data+dumped, 0);
  dumped += dumplang(xml, data+dumped, 1);

  return dumped;
}

void parsetexts(xmlinf *inf)
{
  adddesc(inf, 0);
  adddesc(inf, 1);
}


// ***********************
//  END OF CLASS HANDLERS
// ***********************

//
// Array defining object class-specific data structures and handlers
//
// Entries are object structure description, its size, dump handler, parse handler
//
objclass objclasses[NUMOBJCLASSES] = {
	{ interfacevars, 24, dumpsimple, parsesimple },	// 00 Interfaces
//	{ sfxvars, 12, dumpsfx, parsesfx},		// 01 Sound effects
	{ NULL, -1, NULL, NULL },			// 01 Sound effects
	{ currvars, 12, dumpcurr, parsecurr },		// 02 Currencies
	{ exhfxvars, 40, dumpexhfx, parseexhfx },	// 03 Exhaust effects
	{ simplevars, 6, dumpsimple, parsesimple },	// 04 Cliff faces
	{ watervars, 14, dumpsimple, parsesimple },	// 05 Water
	{ groundvars, 30, dumpground, parseground },	// 06 Ground
	{ townvars, 26, dumptownnames, parsetownnames },// 07 Town names
	{ cargovars, 31, dumpcargo, parsecargo },	// 08 Cargos
	{ fencevars, 10, dumpsimple, parsesimple },	// 09 Fences
	{ signalvars, 30, dumpsignals, parsesignals },	// 0A Signals
	{ crossingvars, 18, dumpsimple, parsesimple },	// 0B Crossings
	{ lightvars, 12, dumpsimple, parsesimple },	// 0C Street lights
	{ simplevars, 6, dumpsimple, parsesimple },	// 0D Tunnels
	{ bridgevars, 44, dumpsimple, parsesimple },	// 0E Bridges
	{ trnstatvars, 174, dumptrnstat, parsetrnstat},	// 0F Train stations
	{ trkmodvars, 18, dumpsimple, parsesimple },	// 10 Track modifications
	{ trackvars, 54, dumptrack, parsetrack },	// 11 Tracks
	{ roadstvars, 110, dumproadst, parseroadst},	// 12 Road stations
	{ trkmodvars, 18, dumpsimple, parsesimple },	// 13 Road modifications
	{ roadvars, 48, dumproad, parseroad },		// 14 Roads
	{ airportvars, 186, dumpairport, parseairport },// 15 Airports
	{ dockvars, 40, dumpdocks, parsedocks },	// 16 Docks
	{ vehvars, 350, dumpveh, parseveh },		// 17 Vehicles
	{ treevars, 76, dumpsimple, parsesimple },	// 18 Trees
	{ simplevars, 6, dumpsimple, parsesimple },	// 19 Snow
	{ climvars, 10, dumpnosprites, parsenosprites },// 1A Climates
	{ shapevars, 14, dumpsimple, parsesimple },	// 1B Shapes
	{ buildingvars, 190, dumpbuildng, parsebuildng},// 1C Buildings
	{ scaffvars, 18, dumpsimple, parsesimple },	// 1D Scaffolding
	{ indvars, 244, dumpind, parseind },		// 1E Industries
	{ regionvars, 18, dumpregion, parseregion },	// 1F Regions
	{ compvars, 56, dumpcomp, parsecomp },		// 20 Companies
	{ simplevars, 6, dumptexts, parsetexts },	// 21 Texts
};
